//g. 2013.
#include <stdio.h>
#include <math.h>
#include <windows.h>
#include <process.h>
#include <mmsystem.h>
#include <mmreg.h>

#pragma comment(lib,     "winmm.lib")
#pragma comment(lib,     "user32.lib")
#pragma comment(lib,     "ole32.lib")

#define PI                 3.141592653589793238
#define OSC_DT             1.0594630943592952645618252949463 //2^(1/12)
#define OSC_ACC_BASE       (6.05 / (float)44100) //tekitou
#define OSC_MAX            16

int srandom()
{
  static int a = 1;
  static int b = 234567;
  static int c = 890123;
  a += b;
  b += c;
  c += a;
  return (a >> 16);
}

float frand()
{
  return float(srandom()) / 32768.0f;
}

struct DelayBuffer
{
  enum
  {
    Max = 100000,
  };
  float Buf[Max];
  unsigned long   Rate;
  int   Index;
  void Fill()
  {
    for(int i = 0 ; i < Rate; i++)
    {
      Buf[i] = frand();
    }
  }
  void Init(unsigned long m)
  {
    Rate = m;
  }
  void Update(float a)
  {
    Buf[Index] = a;
    Index = (Index + 1) % Rate;
  }
  float Sample(unsigned long n = 0)
  {
    return Buf[ (Index + n) % Rate];
  }
};

//http://www.ari-web.com/service/soft/reverb-2.htm
struct Reverb
{
  enum
  {
    CombMax = 4,
    AllMax  = 2,
  };
  DelayBuffer comb[CombMax];
  DelayBuffer all[AllMax];

  float Sample(float a, int index = 0, int character = 0, int lpfnum = 4)
  {
    const int tau[][4][4] = 
    {
      //prime
      {
        {2063, 1847, 1523, 1277},
        {3089, 2927, 2801, 2111},
        {5479, 5077, 4987, 4057},
        {9929, 7411, 4951, 1063},
      },
      {
        {2053, 1867, 1531, 1259},
        {3109, 2939, 2803, 2113},
        {5477, 5059, 4993, 4051},
        {9949, 7393, 4957, 1097},
      }
    };

    const float gain[] = 
    {
       -(0.7733),
       -(0.7623),
       -(0.7513),
       -(0.7403),
    };
    float D = a * 0.1;
    float E = 0;
    
    //Comb
    for(int i = 0 ; i < CombMax; i++)
    {
      DelayBuffer *reb = &comb[i];
      reb->Init(tau[character % 2][index % 4][i]);
      float k = 0;
      float c = 0;
      int LerpMax = lpfnum + 1;
      for(int h = 0 ; h < LerpMax; h++)
      {
        k += reb->Sample(h);
      }
      k /= float(LerpMax);
      c = a + k;
      reb->Update(c * gain[i]);
      E += c;
    }
    
    //all pass
    const int alltau[] = 
    {
       353, 127,
       751, 353,
       353, 157,
      1039, 751,
    };
    
    for(int i = 0 ; i < AllMax; i++)
    {
      DelayBuffer *reb = &all[i];
      reb->Init(alltau[i]);
      float g = 0.7;
      float c = 0;
      int LerpMax = 4;
      for(int h = 0 ; h < LerpMax; h++)
      {
        c += E + reb->Sample(h) * g;
      }
      c /= float(LerpMax);
      reb->Update(c);
      E = c * -g;
    }
    
    D = (D + E) * 0.5;
    return D;
  }
};

struct Osc
{
  double Phase;
  double Acc;
  int    Count;
  double Volume;
  int    Env[4];
  double DVol[4];
  double Buf;
  
  union 
  {
    struct 
    {
      unsigned Type    : 4;
      unsigned Detune  : 2;
      unsigned Fm      : 2;
      unsigned Phase   : 4;
      unsigned Reseved : 4;
      unsigned Env0    : 4;
      unsigned Env1    : 4;
      unsigned Env2    : 4;
      unsigned Env3    : 4;
    };
    unsigned long Raw;
  } Config;
  
  double GetEnv()
  {
    for(int i = 0, ep = 0; i < 4; i++)
    {
      ep += Env[i];
      if(Count < ep)
      {
        return DVol[i];
      }
    }
    Acc = 0;
    Volume = 0;
    return 0;
  }
  
  void Sample(float *lr)
  {
    float vol  = 0;
    if(Acc > 0)
    {
      double env    = GetEnv();
      double td     = 1.0;
      double dph    = 0.3333;
      unsigned char type = Config.Type;
      const int IOsc = 4;
      for(int i = 0 ; i < IOsc; i++)
      {
        dph *= 3;
        double ph  = td * (Phase + dph) * 3.141592 * 2;
        ph *= ((i&1) ? 1 : -1);
        ph  = sin(ph);
        if(type & 0x1) vol += sin(2.72 * ph);
        //if(type & 0x2) vol += ((ph > 0.45) ? -1 : 1) * 0.5;
        td += 0.005;
      }
      Volume += env;
      Phase  += Acc;
      Phase  += frand() * 0.001;
      vol *= Volume;
      Count++;
    }
    Buf *= 0.999;
    
    //lpf
    vol = vol - (vol - Buf) * (1.0 - (Volume * Volume * Volume) );
    Buf = vol;
    lr[0] = vol;
    lr[1] = vol;
  }
  
  void On(int acc, unsigned long cfg)
  {
    Config.Raw = cfg;
    Acc    = OSC_ACC_BASE * OSC_DT;
    //Phase  = 0;
    Count  = 0;
    Volume = 0;
    
    //tekitou
    for(int i = 0 ; i < acc; i++)
    {
      Acc *= OSC_DT;
    }
    
    int    Rate = 0x489;
    const double EnvRate[] = 
    {
      1.0, -0.5, 0.0, -0.5
    };
    
    //tekitou
    Env[0]   = (Config.Env3 * Rate * 1) + Rate;
    Env[1]   = (Config.Env2 * Rate * 2) + Rate;
    Env[2]   = (Config.Env1 * Rate * 3) + Rate;
    Env[3]   = (Config.Env0 * Rate * 3) + Rate;
    
    for(int i = 0; i < 4; i++)
    {
      DVol[i] = EnvRate[i] / double(Env[i]);
    }
  }
};

namespace Sound
{
  typedef float t_snd;
  struct DataType
  {
    float Left, Right;
  };
  
  VOID FillThread(void *p);
  
  VOID CALLBACK waveOutProc(HWAVEOUT hwo, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2);

  //http://msdn.microsoft.com/ja-jp/library/cc352237.aspx
  //http://msdn.microsoft.com/en-us/library/windows/desktop/dd317593(v=vs.85).aspx
  struct Stream
  {
    enum
    {
      Bits         = (sizeof(t_snd) * 8),
      Channel      = (2),
      Freq         = (44100),
      Align        = ((Channel * Bits) / 8),
      BytePerSec   = (Freq * Align),
      BufferNumber = 6,
      Samples      = 512,
      OscMax       = 10,
    };
    Osc osc[OscMax];
    
    //buffer
    char               soundbuffer[BufferNumber][BytePerSec];
    WAVEHDR            whdr[BufferNumber];
    HANDLE             hSem;
    HANDLE             hThread;
    
    DWORD              index;
    HWAVEOUT           hWaveOut;
    CRITICAL_SECTION   wavecs;
    
    //reverb
    Reverb             rebL;
    Reverb             rebR;
    
    Stream()
    {
      memset(this, 0, sizeof(*this));
    }
    
    void Update(void *buffer, int bytes)
    {
      DataType *buf = (DataType *)buffer;
      int count = bytes / sizeof(DataType);
      
      //reverb config
      static int type   = 0;
      static int lpfnum = 4;
      for(int i = 0; i < count; i++)
      {
        DataType d =
        {
          0, 0,
        };
        for(int o = 0; o < OscMax; o++)
        {
          float data[2] = {0, 0};
          osc[o].Sample(data);
          d.Left  += data[0] * 0.5;
          d.Right += data[1] * 0.5;
        }

        //control.
        if(GetAsyncKeyState(VK_UP) & 0x0001)
        {
          lpfnum = (lpfnum  + 1) % 32;
        }
        
        if(GetAsyncKeyState(VK_DOWN) & 0x0001)
        {
          type = (type  + 1) % 4;
        }
        
        if( !(GetAsyncKeyState(VK_F5) & 0x8000))
        {
          d.Left  = rebL.Sample(d.Left,  type, 0, lpfnum);
          d.Right = rebR.Sample(d.Right, type, 1, lpfnum);
        }
        
        //out
        buf[i] = d;
      }
      printf("lpfnum:%d : type=%d : mon : %03.6f\r",lpfnum, type, abs(buf[0].Left));
    }

    void Init(void)
    {
      memset(soundbuffer, 0, sizeof(soundbuffer));
      memset(osc, 0, sizeof(osc));
      index = 0;
      hSem  = NULL;
      hThread = NULL;
      WAVEFORMATEX wfx = { WAVE_FORMAT_IEEE_FLOAT, Channel, Freq, BytePerSec, Align, Bits, 0 };
      InitializeCriticalSection(&wavecs);
      int ret = waveOutOpen(&hWaveOut, WAVE_MAPPER, &wfx, (DWORD)waveOutProc, (DWORD)this, CALLBACK_FUNCTION);
      if(!ret)
      {
        hSem = CreateSemaphore(NULL, BufferNumber - 1, BufferNumber, NULL);
        
        for(int i = 0 ; i < BufferNumber; i++)
        {
          WAVEHDR temp = { soundbuffer[i], Samples * wfx.nBlockAlign, 0, 0, 0, 0, NULL, 0 };
          whdr[i] = temp;
          waveOutPrepareHeader(hWaveOut, &whdr[i], sizeof(WAVEHDR));
          waveOutWrite(hWaveOut, &whdr[i], sizeof(WAVEHDR));
        }
        hThread = (HANDLE)_beginthread(FillThread, 0, this);
      }
    }

    void Term()
    {
      if(hWaveOut)
      {
        int count = 0 ;
        HWAVEOUT hwo = hWaveOut;
        hWaveOut = NULL;
        WaitForSingleObject(hThread, INFINITE);
        CloseHandle(hThread);
        do {
          count = 0;
          for(int i = 0; i < BufferNumber; i++)
            count += (whdr[i].dwFlags & WHDR_DONE) ? 0 : 1;
          if(count) Sleep(50);
        } while(count);
        if(hSem) CloseHandle(hSem);

        for(int i = 0 ; i < BufferNumber ; i++)
          waveOutUnprepareHeader(hwo, &whdr[i], sizeof(WAVEHDR));
        waveOutReset(hwo);
        waveOutClose(hwo);
        DeleteCriticalSection(&wavecs);
      }
    }
  };
  
  VOID FillThread(void *p)
  {
    Stream *stream = (Stream *)p;
    while(stream->hWaveOut)
    {
      EnterCriticalSection(&stream->wavecs);
      DWORD index = stream->index;
      stream->Update((void *)stream->whdr[index].lpData, stream->whdr[index].dwBufferLength);
      waveOutWrite(stream->hWaveOut, &stream->whdr[index], sizeof(WAVEHDR));
      stream->index = (index + 1) % Stream::BufferNumber;
      LeaveCriticalSection(&stream->wavecs);
      WaitForSingleObject(stream->hSem, INFINITE);
    }
  }
  
  VOID CALLBACK waveOutProc(HWAVEOUT hwo, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
  {
    if(uMsg == WOM_DONE) ReleaseSemaphore(((Stream *)dwInstance)->hSem, 1, NULL);
  }
}; //Sound


Sound::Stream stream;

int main(int argc, char *argv[])
{
  CoInitialize(NULL);
  stream.Init();
  int base = 0x30;
  printf("Tone : A/S/D/F Key\nPitch: Left/Right Key\nLPF  : Up\nType : Down\nF5   : NO Reverb\n");
  while( !(GetAsyncKeyState(VK_ESCAPE) & 0x8000))
  {
    if(GetAsyncKeyState('A') & 0x8000)stream.osc[0].On(base + 0 - 12 ,0x020400FF);
    if(GetAsyncKeyState('S') & 0x8000)stream.osc[1].On(base + 4 - 0  ,0x020400FF);
    if(GetAsyncKeyState('D') & 0x8000)stream.osc[2].On(base + 7 - 0  ,0x020400FF);
    if(GetAsyncKeyState('F') & 0x8000)stream.osc[3].On(base + 11- 0  ,0x020400FF);
    if(GetAsyncKeyState(VK_LEFT)  & 0x8000)base--;
    if(GetAsyncKeyState(VK_RIGHT) & 0x8000)base++;
    Sleep(32);
  }
  stream.Term();
  return 0;
}

